#include<libtransistor/err.h>

#define __LINUX_ERRNO_EXTENSIONS__
#include<errno.h>

#include<libtransistor/types.h>
#include<libtransistor/util.h>
#include<libtransistor/svc.h> // for SVC_ID_*
#include<libtransistor/ipc/fs/err.h>

typedef struct {
	result_t code;
	const char *description;
	int closest_errno;
} err_ent_t;

typedef struct {
	const char *name;
	int group_no;
	err_ent_t descriptions[];
} err_group_t;

static err_group_t err_g_official_fs = {
	"FS", -1, {
		{ FSPSRV_ERR_NOT_FOUND, "NOT_FOUND", ENOENT },
		{ FSPSRV_ERR_EXISTS, "EXISTS", EEXIST },
		{ FSPSRV_ERR_DIRECTORY_NOT_EMPTY, "DIRECTORY_NOT_EMPTY", ENOTEMPTY },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_misc = {
	"Misc", 0, {
		{ LIBTRANSISTOR_ERR_UNSPECIFIED, "UNSPECIFIED", ENOSYS },
		{ LIBTRANSISTOR_ERR_UNIMPLEMENTED, "UNIMPLEMENTED", ENOSYS },
		{ LIBTRANSISTOR_ERR_OUT_OF_MEMORY, "OUT_OF_MEMORY", ENOMEM },
		{ LIBTRANSISTOR_ERR_ASSERTION_FAILED, "ASSERTION_FAILED", ENOTRECOVERABLE },
		{ LIBTRANSISTOR_ERR_LEGACY_CONTEXT, "LEGACY_CONTEXT", ENOTSUP },
		{ LIBTRANSISTOR_ERR_MODULE_NOT_INITIALIZED, "MODULE_NOT_INITIALIZED", EAGAIN },
		{ LIBTRANSISTOR_ERR_INVALID_ARGUMENT, "INVALID_ARGUMENT", EINVAL },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_ipc = {
	"IPC", 1, {
		{ LIBTRANSISTOR_ERR_UNSUPPORTED_BUFFER_TYPE, "UNSUPPORTED_BUFFER_TYPE", ENOTSUP },
		{ LIBTRANSISTOR_ERR_TOO_MANY_BUFFERS, "TOO_MANY_BUFFERS", EINVAL },
		{ LIBTRANSISTOR_ERR_INVALID_REQUEST_TYPE, "INVALID_REQUEST_TYPE", EBADR },
		{ LIBTRANSISTOR_ERR_TOO_MANY_HANDLES, "TOO_MANY_HANDLES", EINVAL },
		{ LIBTRANSISTOR_ERR_INVALID_BUFFER_ADDRESS, "INVALID_BUFFER_ADDRESS", EFAULT },
		{ LIBTRANSISTOR_ERR_INVALID_BUFFER_SIZE, "INVALID_BUFFER_SIZE", EINVAL },
		{ LIBTRANSISTOR_ERR_INVALID_PROTECTION, "INVALID_PROTECTION", EINVAL },
		{ LIBTRANSISTOR_ERR_INVALID_IPC_RESPONSE_TYPE, "INVALID_IPC_RESPONSE_TYPE", EBADE },
		{ LIBTRANSISTOR_ERR_INVALID_IPC_RESPONSE_MAGIC, "INVALID_IPC_RESPONSE_MAGIC", EBADE },
		{ LIBTRANSISTOR_ERR_INVALID_RAW_DATA_SIZE, "INVALID_RAW_DATA_SIZE", EBADE },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_RAW_DATA_SIZE, "UNEXPECTED_RAW_DATA_SIZE", EBADE },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_PID, "UNEXPECTED_PID", EBADE },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_COPY_HANDLES, "UNEXPECTED_COPY_HANDLES", EBADE },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_MOVE_HANDLES, "UNEXPECTED_MOVE_HANDLES", EBADE },
		{ LIBTRANSISTOR_ERR_ALREADY_A_DOMAIN, "ALREADY_A_DOMAIN", EINVAL },
		{ LIBTRANSISTOR_ERR_CANT_SEND_OBJECT_ACROSS_DOMAINS, "CANT_SEND_OBJECT_ACROSS_DOMAINS", EINVAL },
		{ LIBTRANSISTOR_ERR_CANT_SEND_DOMAIN_OBJECT_TO_SESSION, "CANT_SEND_DOMAIN_OBJECT_TO_SESSION", EINVAL },
		{ LIBTRANSISTOR_ERR_TOO_MANY_OBJECTS, "TOO_MANY_OBJECTS", EINVAL },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_DOMAIN_HEADER_COMMAND, "UNEXPECTED_DOMAIN_HEADER_COMMAND", EINVAL },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_OBJECTS, "UNEXPECTED_OBJECTS", EINVAL },
		{ LIBTRANSISTOR_ERR_CANT_CLOSE_SESSIONS_LIKE_DOMAIN_OBJECTS, "CANT_CLOSE_SESSIONS_LIKE_DOMAIN_OBJECTS", EINVAL },
		{ LIBTRANSISTOR_ERR_MALFORMED_CLOSE_REQUEST, "MALFORMED_CLOSE_REQUEST", EBADE },
		{ LIBTRANSISTOR_ERR_INVALID_IPC_REQUEST_MAGIC, "INVALID_IPC_REQUEST_MAGIC", EBADE },
		{ LIBTRANSISTOR_ERR_CANT_SEND_SESSION_OBJECT_FROM_DOMAIN, "CANT_SEND_SESSION_OBJECT_FROM_DOMAIN", EINVAL },
		{ LIBTRANSISTOR_ERR_CANT_SEND_DOMAIN_OBJECT_FROM_SESSION, "CANT_SEND_DOMAIN_OBJECT_FROM_SESSION", EINVAL },
		{ LIBTRANSISTOR_ERR_UNEXPECTED_BUFFER_PROTECTION, "UNEXPECTED_BUFFER_PROTECTION", EINVAL },
		{ LIBTRANSISTOR_ERR_REFUSAL_TO_CONVERT_BORROWED_OBJECT, "REFUSAL_TO_CONVERT_BORROWED_OBJECT", EINVAL },
		{ LIBTRANSISTOR_ERR_EXPECTED_SESSION_CLOSURE, "EXPECTED_SESSION_CLOSURE", EBADE },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_sm = {
	"SM", 2, {
		{ LIBTRANSISTOR_ERR_SM_NOT_INITIALIZED, "SM_NOT_INITIALIZED", EBADFD },
		{ LIBTRANSISTOR_ERR_SM_SERVICE_NAME_TOO_LONG, "SM_SERVICE_NAME_TOO_LONG", EINVAL },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_bsd = {
	"BSD", 3, {
		{ LIBTRANSISTOR_ERR_BSD_ERRNO_SET, "BSD_ERRNO_SET", ENOSYS },
		{ LIBTRANSISTOR_ERR_BSD_BUFFER_TOO_SMALL, "BSD_BUFFER_TOO_SMALL", ENOMEM },
		{ LIBTRANSISTOR_ERR_BSD_INVALID_MAGIC, "BSD_INVALID_MAGIC", EIO },
		{ LIBTRANSISTOR_ERR_BSD_UNRECOGNIZED_SOCKET_SERVICE, "BSD_UNRECOGNIZED_SOCKET_SERVICE", ENOSYS },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_nv = {
	"NV", 4, {
		{ LIBTRANSISTOR_ERR_NV_INITIALIZE_FAILED, "NV_INITIALIZE_FAILED", EBADE },
		{ LIBTRANSISTOR_ERR_NV_OPEN_FAILED, "NV_OPEN_FAILED", EBADE },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_parcel_and_display = {
	"Parcel & Display", 5, {
		{ LIBTRANSISTOR_ERR_PARCEL_INVALID_BINDER_TYPE, "PARCEL_INVALID_BINDER_TYPE", EINVAL },
		{ LIBTRANSISTOR_ERR_PARCEL_DATA_TOO_BIG, "PARCEL_DATA_TOO_BIG", EINVAL },
		{ LIBTRANSISTOR_ERR_PARCEL_DATA_UNDERRUN, "PARCEL_DATA_UNDERRUN", ENODATA },
		{ LIBTRANSISTOR_ERR_DISPLAY_FENCE_TOO_MANY_FDS, "DISPLAY_FENCE_TOO_MANY_FDS", EINVAL },
		{ LIBTRANSISTOR_ERR_DISPLAY_GRAPHIC_BUFFER_LENGTH_MISMATCH, "DISPLAY_GRAPHIC_BUFFER_LENGTH_MISMATCH", EINVAL },
		{ LIBTRANSISTOR_ERR_SURFACE_CONNECT_FAILED, "SURFACE_CONNECT_FAILED", EBADE },
		{ LIBTRANSISTOR_ERR_SURFACE_DEQUEUE_BUFFER_FAILED, "SURFACE_DEQUEUE_BUFFER_FAILED", EBADE },
		{ LIBTRANSISTOR_ERR_SURFACE_QUEUE_BUFFER_FAILED, "SURFACE_QUEUE_BUFFER_FAILED", EBADE },
		{ LIBTRANSISTOR_ERR_SURFACE_INVALID_STATE, "SURFACE_INVALID_STATE", EBUSY },
		{ LIBTRANSISTOR_ERR_DISPLAY_INVALID_FENCE, "DISPLAY_INVALID_FENCE", EBUSY },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_gpu = {
	"GPU", 6, {
		{ LIBTRANSISTOR_ERR_GPU_BUFFER_UNALIGNED, "GPU_BUFFER_UNALIGNED", EFAULT },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_hid = {
	"HID", 7, {
		{ LIBTRANSISTOR_ERR_HID_BAD_STRUCTURE, "HID_BAD_STRUCTURE", EFAULT },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_fs = {
	"FS", 8, {
		{ LIBTRANSISTOR_ERR_FS_INTERNAL_ERROR, "FS_INTERNAL_ERROR", EINVAL },
		{ LIBTRANSISTOR_ERR_FS_OUT_OF_DIR_ENTRIES, "FS_OUT_OF_DIR_ENTRIES", 0 },
		{ LIBTRANSISTOR_ERR_FS_NAME_TOO_LONG, "FS_NAME_TOO_LONG", ENAMETOOLONG },
		{ LIBTRANSISTOR_ERR_FS_NOT_A_DIRECTORY, "FS_NOT_A_DIRECTORY", ENOTDIR },
		{ LIBTRANSISTOR_ERR_FS_NOT_A_FILE, "FS_NOT_A_FILE", EIO },
		{ LIBTRANSISTOR_ERR_FS_NOT_FOUND, "FS_NOT_FOUND", ENOENT },
		{ LIBTRANSISTOR_ERR_FS_INVALID_PATH, "FS_INVALID_PATH", ENOENT },
		{ LIBTRANSISTOR_ERR_FS_PATH_TOO_DEEP, "FS_PATH_TOO_DEEP", ENOENT },
		{ LIBTRANSISTOR_ERR_FS_PATH_EXISTS, "FS_PATH_EXISTS", EEXIST },
		{ LIBTRANSISTOR_ERR_FS_READ_ONLY, "FS_READ_ONLY", EROFS },
		{ LIBTRANSISTOR_ERR_FS_ACCESS_DENIED, "FS_ACCESS_DENIED", EACCES },
		{ LIBTRANSISTOR_ERR_FS_IO_ERROR, "FS_IO_ERROR", EIO },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_am = {
	"AM", 9, {
		{ LIBTRANSISTOR_ERR_AM_WORKAROUND_ACTIVE, "AM_WORKAROUND_ACTIVE", ENOSYS },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_ipc_server = {
	"IPC Server", 10, {
		{ LIBTRANSISTOR_ERR_IPCSERVER_INVALID_SESSION_STATE, "IPCSERVER_INVALID_SESSION_STATE", EINVAL },
		{ LIBTRANSISTOR_ERR_IPCSERVER_TOO_MANY_SESSIONS, "IPCSERVER_TOO_MANY_SESSIONS", EMFILE },
		{ LIBTRANSISTOR_ERR_IPCSERVER_NO_SUCH_OBJECT, "IPCSERVER_NO_SUCH_OBJECT", ENOENT },
		{ LIBTRANSISTOR_ERR_IPCSERVER_OBJECT_NOT_ACTIVE, "IPCSERVER_OBJECT_NOT_ACTIVE", EINVAL },
		{ LIBTRANSISTOR_ERR_IPCSERVER_SESSION_OBJECT_WAS_DOMAIN, "IPCSERVER_SESSION_OBJECT_WAS_DOMAIN", EINVAL },
		{ LIBTRANSISTOR_ERR_IPCSERVER_CANT_SEND_ROOT_OBJECT, "IPCSERVER_CANT_SEND_ROOT_OBJECT", EINVAL },
		{ LIBTRANSISTOR_ERR_IPCSERVER_NO_SUCH_COMMAND, "IPCSERVER_NO_SUCH_COMMAND", ENOSYS },
		{ LIBTRANSISTOR_ERR_IPCSERVER_TOO_MANY_PORTS, "IPCSERVER_TOO_MANY_PORTS", EMFILE },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_page_allocator = {
	"Page Allocator", 11, {
		{ LIBTRANSISTOR_ERR_AP_OUT_OF_PAGES, "AP_OUT_OF_PAGES", ENOMEM },
		{ LIBTRANSISTOR_ERR_AP_DUPLICATE_BLOCK, "AP_DUPLICATE_BLOCK", ENOMEM },
		{ LIBTRANSISTOR_ERR_AP_INSERT_FAILED, "AP_INSERT_FAILED", ENOMEM },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_transistor_linker = {
	"Transistor Linker", 12, {
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_MODULE_HEADER, "TRNLD_INVALID_MODULE_HEADER", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_DUPLICATE_DT_ENTRY, "TRNLD_DUPLICATE_DT_ENTRY", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_MISSING_DT_ENTRY, "TRNLD_MISSING_DT_ENTRY", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_RELOC_ENT, "TRNLD_INVALID_RELOC_ENT", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_RELOC_TABLE_SIZE, "TRNLD_INVALID_RELOC_TABLE_SIZE", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_RELA_SYMBOL_UNSUPPORTED, "TRNLD_RELA_SYMBOL_UNSUPPORTED", ENOSYS },
		{ LIBTRANSISTOR_ERR_TRNLD_UNRECOGNIZED_RELOC_TYPE, "TRNLD_UNRECOGNIZED_RELOC_TYPE", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_FAILED_TO_READ_MODULE, "TRNLD_FAILED_TO_READ_MODULE", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_FAILED_TO_FIND_MODULE, "TRNLD_FAILED_TO_FIND_MODULE", ELIBACC },
		{ LIBTRANSISTOR_ERR_TRNLD_NOT_EXECUTABLE, "TRNLD_NOT_EXECUTABLE", ELIBEXEC },
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_RELOC_TABLE_TYPE, "TRNLD_INVALID_RELOC_TABLE_TYPE", ENOSYS },
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_SYM_ENT, "TRNLD_INVALID_SYM_ENT", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_NEEDS_STRTAB, "TRNLD_NEEDS_STRTAB", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_NEEDS_SYMTAB, "TRNLD_NEEDS_SYMTAB", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_COULD_NOT_RESOLVE_SYMBOL, "TRNLD_COULD_NOT_RESOLVE_SYMBOL", ENOENT },
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_MODULE_TYPE, "TRNLD_INVALID_MODULE_TYPE", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_INVALID_MODULE_STATE, "TRNLD_INVALID_MODULE_STATE", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_NO_LOADER_FOR_MODULE, "TRNLD_NO_LOADER_FOR_MODULE", ENOSYS },
		{ LIBTRANSISTOR_ERR_TRNLD_MALFORMED_NRO, "TRNLD_MALFORMED_NRO", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_NEEDS_PROCESS_HANDLE, "TRNLD_NEEDS_PROCESS_HANDLE", EINVAL },
		{ LIBTRANSISTOR_ERR_TRNLD_MALFORMED_ELF, "TRNLD_MALFORMED_ELF", ELIBBAD },
		{ LIBTRANSISTOR_ERR_TRNLD_UNSUPPORTED_ELF, "TRNLD_UNSUPPORTED_ELF", ELIBBAD },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_syscalls = {
	"SVC", 13, {
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_HEAP_SIZE), "NEEDS_SET_HEAP_SIZE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_MEMORY_PERMISSION), "NEEDS_SET_MEMORY_PERMISSION", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_MEMORY_ATTRIBUTE), "NEEDS_SET_MEMORY_ATTRIBUTE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_MEMORY), "NEEDS_MAP_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_UNMAP_MEMORY), "NEEDS_UNMAP_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_QUERY_MEMORY), "NEEDS_QUERY_MEMORY", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_EXIT_PROCESS), "NEEDS_EXIT_PROCESS", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_THREAD), "NEEDS_CREATE_THREAD", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_START_THREAD), "NEEDS_START_THREAD", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_EXIT_THREAD), "NEEDS_EXIT_THREAD", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SLEEP_THREAD), "NEEDS_SLEEP_THREAD", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_THREAD_PRIORITY), "NEEDS_GET_THREAD_PRIORITY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_THREAD_PRIORITY), "NEEDS_SET_THREAD_PRIORITY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_THREAD_CORE_MASK), "NEEDS_GET_THREAD_CORE_MASK", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_THREAD_CORE_MASK), "NEEDS_SET_THREAD_CORE_MASK", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_CURRENT_PROCESSOR_NUMBER), "NEEDS_GET_CURRENT_PROCESSOR_NUMBER", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SIGNAL_EVENT), "NEEDS_SIGNAL_EVENT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CLEAR_EVENT), "NEEDS_CLEAR_EVENT", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_SHARED_MEMORY), "NEEDS_MAP_SHARED_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_UNMAP_SHARED_MEMORY), "NEEDS_UNMAP_SHARED_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_TRANSFER_MEMORY), "NEEDS_CREATE_TRANSFER_MEMORY", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CLOSE_HANDLE), "NEEDS_CLOSE_HANDLE", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_RESET_SIGNAL), "NEEDS_RESET_SIGNAL", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_WAIT_SYNCHRONIZATION), "NEEDS_WAIT_SYNCHRONIZATION", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CANCEL_SYNCHRONIZATION), "NEEDS_CANCEL_SYNCHRONIZATION", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_ARBITRATE_LOCK), "NEEDS_ARBITRATE_LOCK", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_ARBITRATE_UNLOCK), "NEEDS_ARBITRATE_UNLOCK", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_WAIT_PROCESS_WIDE_KEY_ATOMIC), "NEEDS_WAIT_PROCESS_WIDE_KEY_ATOMIC", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SIGNAL_PROCESS_WIDE_KEY), "NEEDS_SIGNAL_PROCESS_WIDE_KEY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_SYSTEM_TICK), "NEEDS_GET_SYSTEM_TICK", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CONNECT_TO_NAMED_PORT), "NEEDS_CONNECT_TO_NAMED_PORT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SEND_SYNC_REQUEST_LIGHT), "NEEDS_SEND_SYNC_REQUEST_LIGHT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SEND_SYNC_REQUEST), "NEEDS_SEND_SYNC_REQUEST", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SEND_SYNC_REQUEST_WITH_USER_BUFFER), "NEEDS_SEND_SYNC_REQUEST_WITH_USER_BUFFER", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SEND_ASYNC_REQUEST_WITH_USER_BUFFER), "NEEDS_SEND_ASYNC_REQUEST_WITH_USER_BUFFER", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_PROCESS_ID), "NEEDS_GET_PROCESS_ID", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_THREAD_ID), "NEEDS_GET_THREAD_ID", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_BREAK), "NEEDS_BREAK", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_OUTPUT_DEBUG_STRING), "NEEDS_OUTPUT_DEBUG_STRING", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_RETURN_FROM_EXCEPTION), "NEEDS_RETURN_FROM_EXCEPTION", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_INFO), "NEEDS_GET_INFO", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_FLUSH_ENTIRE_DATA_CACHE), "NEEDS_FLUSH_ENTIRE_DATA_CACHE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_FLUSH_DATA_CACHE), "NEEDS_FLUSH_DATA_CACHE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_LAST_THREAD_INFO), "NEEDS_GET_LAST_THREAD_INFO", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_RESOURCE_LIMIT_LIMIT_VALUE), "NEEDS_GET_RESOURCE_LIMIT_LIMIT_VALUE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_RESOURCE_LIMIT_CURRENT_VALUE), "NEEDS_GET_RESOURCE_LIMIT_CURRENT_VALUE", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_THREAD_ACTIVITY), "NEEDS_SET_THREAD_ACTIVITY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_THREAD_CONTEXT3), "NEEDS_GET_THREAD_CONTEXT3", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_DUMP_INFO), "NEEDS_DUMP_INFO", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_SESSION), "NEEDS_CREATE_SESSION", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_ACCEPT_SESSION), "NEEDS_ACCEPT_SESSION", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_REPLY_AND_RECEIVE_LIGHT), "NEEDS_REPLY_AND_RECEIVE_LIGHT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_REPLY_AND_RECEIVE), "NEEDS_REPLY_AND_RECEIVE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_REPLY_AND_RECEIVE_WITH_USER_BUFFER), "NEEDS_REPLY_AND_RECEIVE_WITH_USER_BUFFER", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_EVENT), "NEEDS_CREATE_EVENT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SLEEP_SYSTEM), "NEEDS_SLEEP_SYSTEM", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_READ_WRITE_REGISTER), "NEEDS_READ_WRITE_REGISTER", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_PROCESS_ACTIVITY), "NEEDS_SET_PROCESS_ACTIVITY", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_SHARED_MEMORY), "NEEDS_CREATE_SHARED_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_TRANSFER_MEMORY), "NEEDS_MAP_TRANSFER_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_UNMAP_TRANSFER_MEMORY), "NEEDS_UNMAP_TRANSFER_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_INTERRUPT_EVENT), "NEEDS_CREATE_INTERRUPT_EVENT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_QUERY_PHYSICAL_ADDRESS), "NEEDS_QUERY_PHYSICAL_ADDRESS", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_QUERY_IO_MAPPING), "NEEDS_QUERY_IO_MAPPING", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_DEVICE_ADDRESS_SPACE), "NEEDS_CREATE_DEVICE_ADDRESS_SPACE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_ATTACH_DEVICE_ADDRESS_SPACE), "NEEDS_ATTACH_DEVICE_ADDRESS_SPACE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_DETACH_DEVICE_ADDRESS_SPACE), "NEEDS_DETACH_DEVICE_ADDRESS_SPACE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_DEVICE_ADDRESS_SPACE_BY_FORCE), "NEEDS_MAP_DEVICE_ADDRESS_SPACE_BY_FORCE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_DEVICE_ADDRESS_SPACE_ALIGNED), "NEEDS_MAP_DEVICE_ADDRESS_SPACE_ALIGNED", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_DEVICE_ADDRESS_SPACE), "NEEDS_MAP_DEVICE_ADDRESS_SPACE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_UNMAP_DEVICE_ADDRESS_SPACE), "NEEDS_UNMAP_DEVICE_ADDRESS_SPACE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_INVALIDATE_PROCESS_DATA_CACHE), "NEEDS_INVALIDATE_PROCESS_DATA_CACHE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_STORE_PROCESS_DATA_CACHE), "NEEDS_STORE_PROCESS_DATA_CACHE", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_FLUSH_PROCESS_DATA_CACHE), "NEEDS_FLUSH_PROCESS_DATA_CACHE", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_DEBUG_ACTIVE_PROCESS), "NEEDS_DEBUG_ACTIVE_PROCESS", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_BREAK_DEBUG_PROCESS), "NEEDS_BREAK_DEBUG_PROCESS", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_TERMINATE_DEBUG_PROCESS), "NEEDS_TERMINATE_DEBUG_PROCESS", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_DEBUG_EVENT), "NEEDS_GET_DEBUG_EVENT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CONTINUE_DEBUG_EVENT), "NEEDS_CONTINUE_DEBUG_EVENT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_PROCESS_LIST), "NEEDS_GET_PROCESS_LIST", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_THREAD_LIST), "NEEDS_GET_THREAD_LIST", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_DEBUG_THREAD_CONTEXT), "NEEDS_GET_DEBUG_THREAD_CONTEXT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_DEBUG_THREAD_CONTEXT), "NEEDS_SET_DEBUG_THREAD_CONTEXT", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_QUERY_DEBUG_PROCESS_MEMORY), "NEEDS_QUERY_DEBUG_PROCESS_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_READ_DEBUG_PROCESS_MEMORY), "NEEDS_READ_DEBUG_PROCESS_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_WRITE_DEBUG_PROCESS_MEMORY), "NEEDS_WRITE_DEBUG_PROCESS_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_HARDWARE_BREAK_POINT), "NEEDS_SET_HARDWARE_BREAK_POINT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_DEBUG_THREAD_PARAM), "NEEDS_GET_DEBUG_THREAD_PARAM", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_PORT), "NEEDS_CREATE_PORT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MANAGE_NAMED_PORT), "NEEDS_MANAGE_NAMED_PORT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CONNECT_TO_PORT), "NEEDS_CONNECT_TO_PORT", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_PROCESS_MEMORY_PERMISSION), "NEEDS_SET_PROCESS_MEMORY_PERMISSION", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_PROCESS_MEMORY), "NEEDS_MAP_PROCESS_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_UNMAP_PROCESS_MEMORY), "NEEDS_UNMAP_PROCESS_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_QUERY_PROCESS_MEMORY), "NEEDS_QUERY_PROCESS_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_MAP_PROCESS_CODE_MEMORY), "NEEDS_MAP_PROCESS_CODE_MEMORY", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_UNMAP_PROCESS_CODE_MEMORY), "NEEDS_UNMAP_PROCESS_CODE_MEMORY", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_PROCESS), "NEEDS_CREATE_PROCESS", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_START_PROCESS), "NEEDS_START_PROCESS", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_TERMINATE_PROCESS), "NEEDS_TERMINATE_PROCESS", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_GET_PROCESS_INFO), "NEEDS_GET_PROCESS_INFO", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CREATE_RESOURCE_LIMIT), "NEEDS_CREATE_RESOURCE_LIMIT", EPERM },
		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_SET_RESOURCE_LIMIT_LIMIT_VALUE), "NEEDS_SET_RESOURCE_LIMIT_LIMIT_VALUE", EPERM },

		{ LIBTRANSISTOR_ERR_NEEDS_SYSCALL(SVC_ID_CALL_SECURE_MONITOR), "NEEDS_CALL_SECURE_MONITOR", EPERM },
		{ 0, NULL, 0 },
	}
};

static err_group_t err_g_usb = {
	"USB", 14, {
		{ LIBTRANSISTOR_ERR_USB_REPORT_NOT_FOUND, "USB_REPORT_NOT_FOUND", ENOENT },
		{ LIBTRANSISTOR_ERR_USB_TRANSFER_FAILED, "USB_TRANSFER_FAILED", EIO },
		{ LIBTRANSISTOR_ERR_USB_ALREADY_BOUND_OTHER_COMPLEX, "USB_ALREADY_BOUND_OTHER_COMPLEX", EINVAL },
		{ 0, NULL, 0 },
	}
};

static err_group_t *err_groups_transistor[] = {
	&err_g_misc,
	&err_g_ipc,
	&err_g_sm,
	&err_g_bsd,
	&err_g_nv,
	&err_g_parcel_and_display,
	&err_g_gpu,
	&err_g_hid,
	&err_g_fs,
	&err_g_am,
	&err_g_ipc_server,
	&err_g_page_allocator,
	&err_g_transistor_linker,
	&err_g_syscalls,
	&err_g_usb,
};

static void lookup_group(result_t code, void *vp_group, trn_result_description_t *out) {
	err_group_t *err_group = vp_group;
	for(size_t code_index = 0; err_group->descriptions[code_index].code != 0; code_index++) {
		if(err_group->descriptions[code_index].code == code) {
			out->description   = err_group->descriptions[code_index].description;
			out->closest_errno = err_group->descriptions[code_index].closest_errno;
			return;
		}
	}
}

static void lookup_transistor(result_t code, void *vp_ignored, trn_result_description_t *out) {
	int group_no = (code >> 9) / 1000;
	out->group = "unknown";
	for(size_t group_index = 0; group_index < ARRAY_LENGTH(err_groups_transistor); group_index++) {
		if(err_groups_transistor[group_index]->group_no == group_no) {
			out->group = err_groups_transistor[group_index]->name;
			lookup_group(code, err_groups_transistor[group_index], out);
			return;
		}
	}
}

typedef void(*module_lookup_function)(result_t code, void *userdata, trn_result_description_t *out);

typedef struct {
	int module;
	const char *name;
	module_lookup_function func;
	void *userdata;
} module_lookup_t;

static module_lookup_t module_lookups[] = {
	{ MODULE_FS, "FS", lookup_group, &err_g_official_fs },
	{ MODULE_LIBTRANSISTOR, "libtransistor", lookup_transistor, NULL },
	{ 0, NULL, NULL },
};

static const trn_result_description_t err_d_ok = { 0, "OK", "OK", "OK", 0};

void trn_lookup_result(result_t code, trn_result_description_t *out) {
	out->code = code;
	out->module = "unknown";
	out->group = "N/A";
	out->description = "unknown";
	out->closest_errno = EINVAL;

	if(code == RESULT_OK) {
		*out = err_d_ok;
		return;
	}

	int module = code & 0x1FF;
	
	for(size_t module_index = 0; module_lookups[module_index].module != 0; module_index++) {
		if(module_lookups[module_index].module == module) {
			out->module = module_lookups[module_index].name;
			module_lookups[module_index].func(code, module_lookups[module_index].userdata, out);
			return;
		}
	}
}
